{-  === Hypers.hs ===

    Framework for writing quantitative information-flow sequential programs
    based on the monadic semantics of [1-3]. Includes definitions of discrete
    distributions, hyper-distributions and program constructors including
    while-loops.
    
    [1] Annabelle McIver, Larissa Meinicke and Carroll Morgan.
        Hidden-Markov program algebra with iteration.
        Mathematical Structures in Computer Science 25,2, pp 320-60.
        10.1017/S0960129513000625
        2015

    [2] Annabelle McIver, Carroll Morgan and Tahiry M. Rabehaja.
        Abstract Hidden Markov Models:
            A monadic account of quantitative information flow.
        Symposium on Logic in Computer Science (LiCS) 30, pp 597-608.
        10.1109/LICS.2015.61
        2015

    [3] Annabelle McIver, Larissa Meinicke and Carroll Morgan.
        Compositional closure for Bayes risk in probabilistic noninterference.
        Automata, Languages and Programming, 37.
        LNCS 6199, pp 223-25.
        10.1007/978-3-642-14162-1_19
        2010
-}

-- Based on an original by Tom Schrijvers, December 2014.
-- Updated and streamlined by Carroll Morgan, October/November 2015.


{-# LANGUAGE 
    TypeSynonymInstances,
    FlexibleInstances,
    ScopedTypeVariables,
    MultiParamTypeClasses,
    FlexibleContexts,
    FunctionalDependencies,
    TupleSections,
    InstanceSigs
  #-}

module Hypers where

import Prelude
import GHC.Exts
import Data.Ratio
import Data.List
import Data.Function
import Control.Monad
import Control.Applicative
import Text.PrettyPrint.Boxes hiding ((<+>))
import qualified Text.PrettyPrint.Boxes

--
-- Basic framework.
--

type Number = Rational
type Prob = Number
newtype Dist a = D { runD_:: [(Prob,a)] } deriving Show -- Dist's are weighted lists.
type Hyper a = Dist(Dist a)

aggregate:: Ord a => [(Prob,a)] -> [(Prob,a)]
aggregate pas = filter ((/=0).fst) $ map sumGroup groupedPairs
  where groupedPairs = groupWith snd $ sortWith snd pas
        sumGroup group = (sum (map fst group), snd (head group))

makeDist:: Ord a => [(Prob,a)] -> Dist a
makeDist pas = D $ aggregate pas
canonDist:: Ord a => Dist a -> Dist a
canonDist = makeDist.runD_

prob:: Eq a => Dist a -> a -> Prob
prob da a  = sum [p | (p,x) <- runD_ da, x==a]


--
-- Monad structure
--

instance Functor Dist where fmap = liftM

-- Can't use makeDist here (for D) because it needs Ord.
instance Monad Dist where
  return x = D [(1, x)]
  m >>= f  = D [ (p*q,y) | (p, x) <- runD_ m, (q, y) <- runD_ (f x) ]
  
-- wiki.haskell.org/Functor-Applicative-Monad_Proposal#Future-proofing_current_code
instance Applicative Dist where
  pure x = D [(1, x)]
  (<*>) = ap


--
-- Utilities on distributions
--

weight:: Dist a -> Prob
weight = sum . map fst . runD_ 

scale:: Rational -> Dist a -> Dist a
scale r da = D [ (r*p,a) | (p,a)<- runD_ da ]

test:: (a->Bool) -> Dist a -> Dist a
test c da = D $ filter (c.snd) $ runD_ da

normalise:: Dist a -> Dist a
normalise da = scale (1/ weight da) da


-- This is so we can write x <|p|> y to make a Dist from atoms.
-- It's analogous to _p\oplus.
-- NOTE 160527: Somehow this ended up backwards; reversed it today, this file only.

infixl 0 <|
infixl 0 |>
(<|):: a -> Prob -> (Prob,a)
x<|p = (p,x)
(|>):: Ord a => (Prob,a) -> a -> Dist a
(p,x)|>y = makeDist [(p,x),(1-p,y)]

class VectorSpace v where
  merge:: [(Number,v)] -> v
  
instance Ord a => VectorSpace (Dist a) where
  merge = makeDist.concat. map scale where
          scale (r,da) = [(r*p,a) | (p,a)<- runD_ da]
        
instance VectorSpace Prob where
  merge ps = sum $ map (\(p,q)->p*q) ps

instance VectorSpace Float where
  merge ps = sum $ map (\(p,q) -> fromRational p*q) ps

instance Ord a => Eq (Dist a) where
  d1 == d2  =  runD_ d1 == runD_ d2
instance Ord a => Ord (Dist a) where
  compare d1 d2  =  compare (runD_ d1) (runD_ d2)

-- This is so we can merge Dist's; it's analogous to _p+.
infixr 0 <<|
infixr 0 |>>
(|>>):: Prob -> Dist a -> (Prob, Dist a)
p|>>db = (p, db)
(<<|):: Ord a => Dist a -> (Prob, Dist a) -> Dist a
da <<| (p, db) = merge [(p,da),(1-p,db)]


-- Here we define our own version of the monad operators, because for efficiency
-- we have to aggregate every time we make a Dist.
--
-- Swapping the commented-out versions for the aggregated ones should typecheck,
-- confirming that we have the right structure.

-- Map (pushforward).
map':: Ord b => (a->b) -> Dist a -> Dist b
map' f da = makeDist [ (p,f a) | (p,a)<- runD_ da ]
-- map' = fmap -- 1304: If this type-checks, then the monadic structure is OK.

-- Join.
join':: Ord a => Hyper a -> Dist a
join' = merge.runD_
-- join' = join -- 1304.

-- Bind.
bind':: Ord b => (a -> Dist b) -> Dist a -> Dist b
bind' f = join'. map' f
-- bind' f da = da >>= f -- 1304.

-- Kleisli.
kleisli':: (Ord b, Ord c) => (b -> Dist c) -> (a -> Dist b) -> (a -> Dist c)
kleisli' f g a = bind' f $ g a
-- kleisli' f g = g >=> f -- 1304.


--
-- Making useful distributions.
--

-- Make a "uniform" distribution of given non-empty support.
-- Repeated elements get higher proportions.
uniform:: Ord a => [a] -> Dist a
-- genericLength gives Integer rather than Int.
uniform l@(x:xs) = makeDist $ map (1%len,) l where len = genericLength l

-- Make a point distribution
point:: a -> Dist a
point a = D [(1,a)] -- No need for aggregation, obviously.

-- Shortcut for equal choice between just two alternatives.
(<+>):: Ord a => a -> a -> Dist a
x <+> y  =  uniform [x,y]

-- Product distributions.
prod:: Dist a -> Dist b -> Dist (a,b) -- 1628: No aggregation needed.
prod da db = D [(p*q,(a,b)) | (p,a)<- runD_ da, (q,b)<- runD_ db]
prod2 = prod
prod3:: Dist a -> Dist b -> Dist c-> Dist (a,b,c) -- 1628.
prod3 da db dc = D [(p*q*r,(a,b,c)) | (p,a)<- runD_ da,
                                      (q,b)<- runD_ db,
                                      (r,c)<- runD_ dc]

--
-- Programs and their constructors
--

type Program a  = Dist a -> Hyper a

-- SKIP program
skip:: Program a
skip = point

-- update converts a markov to a program.
type Markov a = a -> Dist a
update:: Ord a => Markov a -> Program a
update mk  =  point . bind' mk

-- reveal converts a channel to a program
type Channel a b = a -> Dist b
reveal:: (Ord a, Ord b) => Channel a b -> Program a
reveal f da = hyper
  where 
    distsByA = [(p,(a,f a)) | (p,a)<- runD_ da]
    joint = [ ((p*q, a), b) | (p,(a,db))<- distsByA, (q,b)<- runD_ db]
    groupedJoint = groupWith snd joint
    removeBs = map (map fst) groupedJoint
    makeInner column = (weight,normed)
      where weight = sum $ map fst column
            normed = makeDist [ (p/weight,a) | (p,a)<- column ]
    hyper = makeDist$ map makeInner removeBs

-- Sequential composition (infix version of Kleisli composition).
(##):: Ord a => Program a -> Program a -> Program a
pa ## pb = kleisli' pb pa

-- External probabilistic choice some function of the state.
extProb:: Ord a => (a -> Prob) -> Program a -> Program a -> Program a
extProb a2p prL prR da =
  let pas = runD_ da
      expanded = concat [[ (p * a2p a,(a,True)),(p * (1-a2p a),(a,False))] | (p,a)<- pas]
      leftList = [(p,a) | (p,(a,b))<- expanded, b]
      rightList = [(p,a) | (p,(a,b))<- expanded, not b]
      leftLen = length leftList
      leftWeight = sum $ map fst leftList
      rightLen = length rightList
      leftDist = normalise $ makeDist leftList
      rightDist = normalise $ makeDist rightList
      leftHyper = prL leftDist
      rightHyper = prR rightDist
   in if      leftLen==0 then  rightHyper
      else if rightLen==0 then leftHyper
      else                     leftHyper <<|leftWeight|>> rightHyper

-- External probabilistic choice constant p is prL <<<|p|>>> prR.
infixr 0 <<<|
infixr 0 |>>>
(|>>>):: Prob -> Program a -> (Prob, Program a)
p|>>>prR = (p, prR)
(<<<|):: Ord a => Program a -> (Prob, Program a) -> Program a
prL<<<|(p, prR) = extProb (const p) prL prR

-- Conditional choice between two programs on condition C is special case of prob choice.
ifC:: Ord a => (a->Bool) -> Program a -> Program a -> Program a
ifC c prL prR = extProb (\a-> if c a then 1 else 0) prL prR
     
-- Repeat a program M times.
repeatM:: Ord a => Int -> Program a -> Program a
repeatM 0 pa da  =  skip da
repeatM n pa da  =  bind' (repeatM (n-1) pa) (pa da) 

-- While-loop based on a condition C.
whileC:: Ord a => (a->Bool) -> Program a -> Program a
whileC c pa = ifC c (pa ## whileC c pa) skip

-- Atomic: No perfect recall, no PC tracking, no reveals honoured.
atomic:: Ord a => Program a -> Program a
atomic pa da = point $ join' $ pa da

-- Local variable: introduce a local variable of type b and its initialisation.
local:: (Ord a,Ord b) => (a -> Dist b) -> Program (a,b) -> Program a
local init pab da = dda
  where expand a = map' ((a,)) $ init a
        dab = bind' expand da
        ddab = pab dab
        dda = map' (map' fst) ddab
        
-- Assertion: Crash program if it could be false.
assert:: (a -> Bool) -> Program a
assert f da = case flip prob True $ map' f da of { 1 -> skip da } 
  
        
--
-- Approximators.
--

discardSmall:: Ord a => Integer -> Hyper a -> Hyper a -- Keep only inners wprob at least 1/n.
discardSmall n dda = makeDist [(p,i) | (p,i)<- runD_ dda, p>=1%n]

discardMany:: Ord a => Int -> Hyper a -> Hyper a -- Keep only n largest inners.
discardMany n dda = makeDist $ take n $ sortWith ((0-).fst) $ runD_ dda 

makeSimplerDist:: Ord a => Integer -> [(Prob,a)] -> Dist a -- Round to nearest simple value.
makeSimplerDist n da = D [ (approxRational p (1%n),a) | (p,a)<- da ]

--
-- Pretty-printing, using Text.PrettyPrint.Boxes.
--

approxDist:: Ord a => Integer -> Dist a -> Dist a
approxDist n dda = makeDist $ map (\(p,da) -> (approxRational p (1%n),da)) $ runD_ dda
approxHyper:: Ord a => Integer -> Hyper a -> Hyper a
approxHyper n = approxDist n . map' (approxDist n)
xHyp:: Ord a => Hyper a -> Hyper a
xHyp = approxHyper 100

pretty:: Pretty a => a -> IO()
pretty t = putStr (render (prettyDoc t))

-- Definition of pretty-printable class.
class Pretty t where prettyDoc:: t -> Box

-- Print Booleans.
instance Pretty Bool where
  prettyDoc False = char '0'
  prettyDoc True  = char '1'

-- Print integers.
instance Pretty Integer where
  prettyDoc n = text $ show n

-- Print strings.
instance Pretty [Char] where
  prettyDoc cs = text $ show cs

-- Print rationals.
instance Pretty Rational where
  prettyDoc r = 
    let ar = r -- approxRational r (1%1000000) -- Avoid huge integers in the printout.
        ar' = if ar==0 then r else ar -- But keep really small fractions.
        n = numerator ar'
        d = denominator ar'
    in if n == 0
         then char '0'
         else if d == 1
                 then text (show n)
                 else  text (show n) <> char '/' <> text (show d)
                 
-- Print floats.
instance Pretty Float where
  prettyDoc f = text $ show f 

-- Print pairs.
instance (Pretty a, Pretty b) => Pretty (a, b) where
  prettyDoc (x,y) = prettyDoc x Text.PrettyPrint.Boxes.<+> prettyDoc y
  
-- Print triples.
instance (Pretty a, Pretty b, Pretty c) => Pretty (a,b,c) where
  prettyDoc (x,y,z) = prettyDoc x Text.PrettyPrint.Boxes.<+>
                      prettyDoc y Text.PrettyPrint.Boxes.<+>
                      prettyDoc z
  
-- Print lists.
instance (Pretty a) => Pretty [a] where
  prettyDoc = foldr (\x xs -> prettyDoc x Text.PrettyPrint.Boxes.<+> xs) nullBox
  
-- Print distributions.
instance (Ord a, Pretty a) => Pretty (Dist a) where
  prettyDoc d =
    vcat left els <> text "   " <> vcat left sps
    where l    = runD_ d
          ps   = map (prettyDoc . fst) l
          els  = map (prettyDoc . snd) l
          sps  = zipWith stretch (map rows els) ps
          stretch r ps  =  ps // emptyBox (r-1) 0

pr:: (Ord a,Pretty a) => Hyper a -> IO()
pr = pretty
px:: (Ord a,Pretty a) => Hyper a -> IO()
px = pretty . approxHyper 100

-- A typical run is 
--		pr$ program prior
-- or	px$ program prior


--
-- Uncertainty loss-function calculators
--

-- Bayes vulnerability: the probability of choosing correctly in one guess.
bayesVulnerability:: Dist a -> Float
bayesVulnerability da = fromRational$ maximum (map fst $ runD_ da)
bayesCoVulnerability da = fromRational $ minimum (map fst $ runD_ da)
bV = bayesVulnerability

-- Classical Shannon entropy
shannon:: Dist a -> Float
shannon da = sum $ map (\(p,_) -> let f = fromRational p in -logBase 2 f * f) $ runD_ da

-- The Kantorovich distance between two discrete distributions
-- on a base type with the discrete metric.
-- Equals half the Manhattan distance.
kantorovich:: Ord a => Dist a -> Dist a -> Float
kantorovich da db =
    let groups = groupWith snd $ sortWith snd (runD_ da ++ runD_ db)
        dist:: [(Prob,a)] -> Float
        dist [(p,_)] = fromRational p
        dist [(p0,_),(p1,_)] = abs $ fromRational (p0-p1) -- snd's should be equal.

    in  (sum $ map dist groups) /2.0 -- Kanto is half the Manhattan.

-- Applies to any of the above, so that they can be used with hypers.
conditional :: (Dist a -> Float) -> Hyper a -> Float
conditional uncertainty = merge . runD_ . map' uncertainty
