{-  Uses hypers.hs, an implementation of monadic quantitative information flow semantics
    in Haskell, to implement an obfuscating log-time exponential program as used for
    public-key cryptography. The program is instrumented with a side-channel leak, and
    produces hyper-distributions showing the effect the leak has had.
    
    Uses LeakyExponential.hs to define an obfuscating log-time exponential program
    that suffers side-channel leakage of conditionals.
    
    Produces tables of leakages of various kinds.
    
    Carroll Morgan, May 2016.
    
-}

-- The hypers.hs module is based on an original by Tom Schrijvers, December 2014,
-- updated and streamlined by Carroll Morgan, October/November 2015.


import Hypers
import LeakyExponential

-- Rows and columns taken from the tables in the FM2016 paper.
table= [(expBits,divisors)| expBits<- [4,5,6,7,8], divisors<-[[2],[2,3],[2,3,5]]]

makeLogBVLeakage:: (Integer,[Integer]) -> Float
makeLogBVLeakage (expBits,divisors) =
    let h= hyper 11 1 divisors 2 [0..2^expBits-1]
        bVAfter = (log $ conditional bV h)/log 2
        bVBefore = fromIntegral $ -expBits
    in bVAfter-bVBefore

makeCoBVLeakage:: (Integer,[Integer]) -> Float
makeCoBVLeakage (expBits,divisors) =
    let h= hyper 11 1 divisors 2 [0..2^expBits-1]
        coBV:: Dist a -> Float
        coBV da = 1.0 - bV da
        coBVAfter = conditional coBV h
        coBVBefore = coBV $ uniform [0..2^expBits-1]
    in coBVBefore-coBVAfter

makeShannonLeakage:: (Integer,[Integer]) -> Float
makeShannonLeakage (expBits,divisors) =
    let h= hyper 11 1 divisors 2 [0..2^expBits-1]
        shannonAfter = conditional shannon h
        shannonBefore = fromIntegral expBits
    in shannonBefore-shannonAfter

-- The first argument is a normaliser, for loss functions more than 1.
makeKantoDistance:: (Integer,[Integer]) -> Float
makeKantoDistance (expBits,divisors) =
    let h= hyper 11 1 divisors 2 [0..2^expBits-1]
        prior= uniform [0..2^expBits-1]
        getE (_,e,_) = e
        hE = map' (map' getE) h
    in  conditional (kantorovich prior) hE
