{-  Uses hypers.hs, an implementation of monadic quantitative information flow semantics
    in Haskell, to implement an obfuscating log-time exponential program as used for
    public-key cryptography. The program is instrumented with a side-channel leak, and
    produces hyper-distributions showing the effect the leak has had.
    
    Carroll Morgan, November 2015.
    
-}

-- The hypers.hs module is based on an original by Tom Schrijvers, December 2014,
-- updated and streamlined by Carroll Morgan, October/November 2015.

-- For non-Haskell readers: Haskell comments are either -- to the end of the line,
-- or over multiple lines between {- and -}.


--
-- Leaky exponential program.
--

{- The program in pseudocode: all operations mod mm.

    VAR bb,ee,pp        -- Base bb, exponent ee, power pp.
    CONST dd,mm         -- List of divisors dd, group modulus
    
    |[  VAR b,e,d:= bb,ee,0                             -- 1: initMarkov
        pp:= 1                                          -- 2: initialise pp
        WHILE e!=0 DO                                   -- 3: notZero
            d<- uniform dd                              -- 4: choose d
            IF e notDivisibleBy d                       -- 5: divisibility check
                THEN pp,e:= pp*b^(e mod d),e-(e mod d)  -- 6: divisibility adjustment
            e,b:= e/d,b^d                               -- 7: division
        DONE
    |]
    -- pp is now ee^bb  
    
-}

-- The above is translated by hand into the Haskell functions below.


module LeakyExponential where

import Hypers   -- Package implementing hyper-distribution monadic semantics.
import Control.Monad
import Text.PrettyPrint.Boxes hiding ((<+>))

-- State space(s) 
type Outer = (Integer,Integer,Integer)  -- Globals: Base bb, exponent ee, power pp.
type Inner = (Integer,Integer,Integer)  -- Locals: Base b, exponent e, divisor d.
type X = (Outer,Inner)                  -- Globals + Locals.

-- Parametrise by atomicity aa, modulus mm, divisor list dd.
--      Atomicity is 00,01,10,11 indicating whether resp WHILE, IF
--        leak is enabled 1 or disabled 0.
--      Modulus is arithmetic group for all the arithmetic.
--      Divisor list selected from uniformly.

program:: Integer -> Integer -> [Integer] -> Program Outer
program aa mm dd = wholeProgram where

-- Primitive elements: Tests, channels and markovs.
  initMarkov:: Outer -> Dist Inner                                  --1 in pseudocode.
  initMarkov (bb,ee,pp) = point (bb,ee,0)

  initPpMarkov:: Markov X
  initPpMarkov ((bb,ee,pp),(b,e,d)) = point ((bb,ee,1),(b,e,d))     --2.

  zeroE:: X -> Bool                                                 --3.
  zeroE (_,(_,e,_)) = e==0

  chooseDMarkov:: Markov X                                          --4.
  chooseDMarkov ((bb,ee,pp),(b,e,d)) = uniform [((bb,ee,pp),(b,e,d)) | d<- dd]
  
  notDivisibleBy:: X -> Bool                                        --5.
  notDivisibleBy (_,(_,e,d))= (e`mod`d)/=0
  
  adjustMarkov:: Markov X                                           --6.
  adjustMarkov ((bb,ee,pp),(b,e,d)) = point ((bb,ee,(pp*b^(e`mod`d))`mod`mm),(b,e-e`mod`d,d))

  divideMarkov:: Markov X                                           --7.
  divideMarkov ((bb,ee,pp),(b,e,d)) = point ((bb,ee,pp),(b^d`mod`mm,e`div`d,d))

-- Program Atoms.
  initPp:: Program X
  initPp = update initPpMarkov          --2.

  chooseD:: Program X
  chooseD = update chooseDMarkov        --4.
  
  adjust:: Program X
  adjust = update adjustMarkov          --6.

  divide:: Program X
  divide = update divideMarkov          --7.

-- IF statement.
  ifS:: Program X
  ifS = ifC notDivisibleBy adjust skip  --5,6.
  
-- WHILE body
  whileB:: Program X
  whileB = chooseD ## ifS ## divide     --4,IF,7.

-- WHILE statement
-- aa is 11 for hide none, 10 for hidden IF, 01 for hidden loop, 00 for hidden all.
  whileS:: Program X
  whileS =                              --4,IF,7.
    case aa of
         11 -> whileC (not.zeroE) $ whileB                          -- leak both
         01 -> (repeatM 40 $ chooseD ## whileB) ## assert zeroE     -- leak IF
         10 -> whileC (not.zeroE) $ atomic $ whileB                 -- leak WHILE
         00 -> (repeatM 40 $ atomic $ chooseD ## ifS) ## assert zeroE   -- leak neither

-- Block body
  blockBody:: Program X
  blockBody = initPp ## whileS -- ## reveal (\((_,_,pp),_)->point pp) --2,WHILE.

-- Block as a whole
  block:: Program Outer
  block = local initMarkov blockBody --1,2,WHILE.
  
-- Whole program
  wholeProgram = block

-- Starting state
prior:: Integer -> [Integer] -> Dist Outer
prior bb ees = uniform $ map (\ee->(bb,ee,0)) ees -- Set pp to 0 arbitrarily.

--
-- Run program
-- Output format is exponent,power innerProb outerProb
--

--      aa         mm         dd           bb         ees
hyper:: Integer -> Integer -> [Integer] -> Integer -> [Integer] -> Hyper Outer
hyper aa mm dd bb ees = program aa mm dd (prior bb ees)

run:: Integer -> Integer -> [Integer] -> Integer -> [Integer] -> IO ()
run aa mm dd bb ees = pretty$ hyper aa mm dd bb ees

--
-- Example: reveal everything, modulus 17, divisors 2,3,5, base 2, exponents 0..15.
-- 

example = run 11 17 [2,3,5] 2 [0..15]

-- To focus on leaks, in a small demo, set mm,bb,ees:= 1,0
-- and project onto ees only.

ex n dd = -- How many exponents n, list of divisors dd.
  let h= hyper 11 1 dd 0 [0..n-1] in
    (pretty $ liftM (liftM (\(a,b,c)->b)) h)
    >>
    putStr ("\nProduced " ++ show (length$ runD_ h) ++ " inners.\n")
