--
-- Two-Boolean XS example from LiCS 2015
--

{-
    Annabelle McIver, Carroll Morgan and Tahiry M. Rabehaja.
    Abstract Hidden Markov Models:
        A monadic account of quantitative information flow.
    Symposium on Logic in Computer Science (LiCS) 30, pp 597-608.
    10.1109/LICS.2015.61
    2015
-}

-- Based on an original by Tom Schrijvers, December 2014.
-- Updated and streamlined by Carroll Morgan, October/November 2015.


import Hypers

-- State space

type XS = (Bool,Bool)

-- Atomic components: two channels and two markovs.

leakLSB:: Channel XS Bool
leakLSB (x0,x1) = point x0

leakMSB:: Channel XS Bool
leakMSB (x0,x1) = point x1

leakOneBit:: Channel XS Bool
leakOneBit (x0,x1) = uniform [x0,x1]

maybeInvertBits:: Markov XS
maybeInvertBits (x0,x1) = uniform [(x0,x1),(not x0,not x1)]

maybeSwapBits:: Markov XS
maybeSwapBits (x0,x1) = uniform [(x0,x1),(x1,x0)]

-- Programs

program0L:: Program XS
program0L = reveal leakLSB                                  -- reveal x0

program0M:: Program XS
program0M = reveal leakMSB                                  -- reveal x1

program0:: Program XS
program0 = reveal leakMSB <<<|1/2|>>> reveal leakLSB        -- reveal x0 [1/2] reveal x1

program1:: Program XS
program1 = reveal leakOneBit                                -- reveal x0[1/2]x1

program2:: Program XS
program2 =    reveal leakOneBit                             -- reveal x0[1/2]x1
           ## update maybeInvertBits                        -- (x0,x1):= (x0,x1)[1/2](-x0,-x1)

program3:: Int -> Program XS
program3 n = repeatM n $ reveal leakOneBit                  -- for i:= 0 to n do
                                                            --   reveal x0[1/2]x1
                                                            -- done
program4:: Int -> Program XS
program4 n =    (repeatM n $ reveal leakOneBit)             -- for i:= 0 to n do
             ## update maybeInvertBits                      --   reveal x0[1/2]x1
                                                            -- done
                                                            -- (x0,x1):= (x0,x1)[1/2](-x0,-x1)
program5:: Program XS
program5 =    reveal leakOneBit                             -- reveal x0[1/2]x1
           ## update maybeSwapBits                          -- (x0,x1):= (x0,x1)[1/2](x1,-x0)
           
program6:: Int -> Program XS
program6 n =    (repeatM n $ reveal leakOneBit)             -- for i:= 0 to n do
             ## update maybeSwapBits                        --   reveal x0[1/2]x1
                                                            -- done
                                                            -- (x0,x1):= (x0,x1)[1/2](x1,x0)

-- Priors

xs = [(False,False), (False,True), (True,False), (True,True)]
u = uniform xs
xs3 = [(False,True), (True,False), (True,True)]
k = uniform xs3

{- These defined in Hypers.hs.
-- Printing

pr:: (Ord a,Pretty a) => Hyper a -> IO()
pr = pretty
px:: (Ord a,Pretty a) => Hyper a -> IO()
px = pretty . approxHyper 100
-}

-- Typical runs, mostly on uniform prior over whole of Boolean^2.
run0L  = pr$ program0L u               
run0M  = pr$ program0M u               
run0   = pr$ program0 u               
run1u  = pr$ program1 u
run1   = run1u            
run1k  = pr$ program1 k
run2   = pr$ program2 u
run3   = pr$ program2 k
run4   = pr$ program3 10 u
run5   = pr$ program4 10 u
run6   = pr$ program3 100 u
run7   = pr$ program4 100 u
run8   = px$ program3 100 u
run9   = px$ program4 100 u
